/* hw_volume_manager.cc
 * This file belongs to Worker, a file manager for UN*X/X11.
 * Copyright (C) 2009 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "hw_volume_manager.hh"
#include "hw_volume.hh"

#ifdef HAVE_HAL_DBUS
#include "hal/libhal-storage.h"
#endif

#include <iostream>
#include <fstream>
#include <sstream>
#include <iterator>
#include <vector>
#include <set>
#include <algorithm>
#include <cctype>

#include "nwc_fsentry.hh"
#include <aguix/util.h>
#include "flagreplacer.hh"
#include "execlass.h"
#include "worker_locale.h"
#include "mounttable_entry.hh"

HWVolumeManager::HWVolumeManager( const std::string &fstab_location,
                                  const std::string &mtab_location ) :
#ifdef HAVE_HAL_DBUS
    m_dbus_conn( NULL ),
    m_hal_ctx( NULL ),
#endif
    m_fstab_table( fstab_location ),
    m_mtab_table( mtab_location ),
    m_mount_command( HW_VM_MOUNT_CMD ),
    m_unmount_command( HW_VM_UNMOUNT_CMD ),
    m_partition_file( HW_VM_PARTS )
{
    HALconnectionInit();
}

HWVolumeManager::~HWVolumeManager()
{
    HALconnectionFini();
}

typedef std::pair< mode_t, dev_t > w_osdevice_t;

static w_osdevice_t getOSDevice( const std::string &dev )
{
    NWC::FSEntry fe( dev );

    if ( ! fe.entryExists() ) {
        return std::make_pair( 0, 0 );
    }

    if ( fe.isLink() ) {
        return std::make_pair( fe.stat_dest_mode(),
                               fe.stat_dest_rdev() );
    } else {
        return std::make_pair( fe.stat_mode(),
                               fe.stat_rdev() );
    }
}

std::list<HWVolume> HWVolumeManager::getVolumes()
{
    std::list<HWVolume> l;
#ifdef HAVE_HAL_DBUS
    DBusError error;
    int i;
    int num_volumes;
    char **volumes;
    const char *udi;
#endif

    m_fstab_table.update();
    m_mtab_table.update();
    m_avail_parts = getAvailablePartitions();
    std::set< w_osdevice_t > managed;

    std::list< MountTableEntry > fstab_entries = m_fstab_table.getEntries();
    for ( std::list< MountTableEntry >::const_iterator it1 = fstab_entries.begin();
          it1 != fstab_entries.end();
          it1++ ) {
        if ( isDevice( it1->getDevice() ) ) {
            //TODO check for duplicate entries?
            HWVolume v;
            v.setDevice( it1->getDevice() );
            //TODO check whether supposed mount point is absolute and exists?
            v.setSupposedMountPoint( it1->getMountPoint() );

            update( v );

            managed.insert( getOSDevice( v.getDevice() ) );
            l.push_back( v );
        }
    }

#ifdef HAVE_HAL_DBUS
    if ( HALreconnect() == 0 ) {

        dbus_error_init( &error );
        volumes = libhal_find_device_by_capability( m_hal_ctx, "volume",
                                                    &num_volumes, &error );
        if ( volumes != NULL ) {
            for (i = 0; i < num_volumes; i++) {
                udi = volumes[i];

                bool use_volume = false;

                // first check whether we should ignore the volume
                try {
                    if ( ! getBoolProperty( udi, "volume.ignore" ) ) {
                        use_volume = true;
                    }
                } catch ( int ) {
                }

                // now check whether the volume is actually a mountable filesystem
                if ( use_volume ) {
                    try {
                        if ( getStringProperty( udi, "volume.fsusage" ) != "filesystem" ) {
                            use_volume = false;
                        }
                    } catch ( int ) {
                    }
                }

                if ( use_volume ) {
                    HWVolume v( udi );

                    update( v );

                    if ( managed.count( getOSDevice( v.getDevice() ) ) < 1 ) {
                        managed.insert( getOSDevice( v.getDevice() ) );
                
                        l.push_back( v );
                    }
                }
            }
            libhal_free_string_array( volumes );
        } else {
            if ( dbus_error_is_set( &error ) ) {
                fprintf( stderr, "HAL Find Volumes Error: %s\n", error.message );
                dbus_error_free( &error );
            }
        }
    }
#endif

#ifndef HAVE_HAL_DBUS
    for ( std::list< std::string >::iterator it1 = m_avail_parts.begin();
          it1 != m_avail_parts.end();
          it1++ ) {
        if ( isDevice( *it1 ) &&
             ! m_fstab_table.containsDevice( *it1 ) &&
             managed.count( getOSDevice( *it1 ) ) < 1 ) {
            HWVolume v;
            v.setDevice( *it1 );
            v.setIsAvailable( true );

            l.push_back( v );
        }
    }
#endif

    return l;
}

int HWVolumeManager::mount( HWVolume &vol, std::string &error_return )
{
    if ( vol.isMounted() == true ) return 0;

    if ( ! vol.getUDI().empty() && vol.isInFStab() == false ) {
        return halMount( vol, error_return );
    } else {
        return cmdMount( vol, error_return );
    }
}

int HWVolumeManager::umount( HWVolume &vol, std::string &error_return )
{
    if ( vol.isMounted() == false ) return 0;

    if ( ! vol.getUDI().empty() && vol.isInFStab() == false ) {
        return halUnmount( vol, error_return );
    } else {
        return cmdUnmount( vol, error_return );
    }
}

void HWVolumeManager::update( HWVolume &vol )
{
    if ( ! vol.getUDI().empty() ) {
        bool mounted = false;
        
        try {
            mounted = getBoolProperty( vol.getUDI(),
                                       "volume.is_mounted" );
        } catch ( int ) {
        }

        if ( mounted ) {
            vol.setIsMounted( true );

            std::string mountpoint;

            try  {
                mountpoint = getStringProperty( vol.getUDI().c_str(),
                                                "volume.mount_point" );
            } catch ( int ) {
            }

            if ( ! mountpoint.empty() ) {
                vol.setMountPoint( mountpoint );
            }
        } else {
            vol.setIsMounted( false );
            vol.setMountPoint( "" );
        }
        
        std::string device;
        try {
            device = getStringProperty( vol.getUDI().c_str(),
                                        "block.device" );
        } catch ( int ) {
        }

        if ( ! device.empty() ) {
            vol.setDevice( device );
            vol.setIsAvailable( true );
        } else {
            vol.setDevice( "" );
            vol.setIsAvailable( false );
        }

        if ( m_fstab_table.containsDevice( vol.getDevice() ) ) {
            vol.setIsInFStab( true );
        } else {
            vol.setIsInFStab( false );
        }
    } else {
        MountTableEntry mte;
        if ( m_fstab_table.containsDevice( vol.getDevice() ) ) {
            vol.setIsInFStab( true );
        } else {
            vol.setIsInFStab( false );
        }
        if ( m_mtab_table.findDevice( vol.getDevice(), mte ) ) {
            vol.setIsMounted( true );
            vol.setIsAvailable( true );
            vol.setMountPoint( mte.getMountPoint() );
        } else {
            vol.setIsMounted( false );
            vol.setMountPoint( "" );
            if ( isAvailablePartition( vol.getDevice(), m_avail_parts ) ) {
                vol.setIsAvailable( true );
            } else {
                vol.setIsAvailable( false );
            }
        }
    }
}

static std::vector< std::string > split( const std::string &str )
{
    std::vector< std::string > l;
    std::string s1;
    std::string::size_type p = 0;

    while ( p < str.length() ) {
        if ( std::isspace( str[p] ) ) {
            if ( ! s1.empty() ) {
                l.push_back( s1 );
                s1 = "";
            }
        } else {
            s1 += str[p];
        }
        p++;
    }
    if ( ! s1.empty() ) {
        l.push_back( s1 );
    }
    return l;
}

std::list< std::string > HWVolumeManager::getAvailablePartitions()
{
    std::list< std::string > l;

    std::ifstream ifile( m_partition_file.c_str() );
    std::string line;
    
    if ( ifile.is_open() ) {
        std::string key;
        int line_nr = 0;
        while ( std::getline( ifile, line ) ) {
            if ( line_nr > 1 && ! line.empty() ) {
                std::vector< std::string > sp = split( line );
                if ( sp.size() == 4 ) {
                    if ( atoi( sp[2].c_str() ) > 10 &&
                         sp[3].length() > 0 ) {
                        char c = sp[3][sp[3].length()-1];
                        if ( c >= '0' && c <= '9' ) {
                            l.push_back( std::string( "/dev/" ) + sp[3] );
                        }
                    }
                }
            }
            line_nr++;
        }
    }

    return l;
}

bool HWVolumeManager::isAvailablePartition( const std::string &dev,
                                            const std::list< std::string > &l )
{
    std::list< std::string >::const_iterator it1;

    for ( it1 = l.begin();
          it1 != l.end();
          it1++ ) {
        if ( isSameDevice( dev, *it1 ) ) return true;
    }
    return false;
}

bool HWVolumeManager::isSameDevice( const std::string &dev1,
                                    const std::string &dev2 )
{
    if ( dev1 == dev2 ) return true;

    w_osdevice_t d1, d2;

    d1 = getOSDevice( dev1 );
    d2 = getOSDevice( dev2 );
    
    if ( d1.first == 0 ||
         d1.second == 0 ) return false;
    if ( d2.first == 0 ||
         d2.second == 0 ) return false;

    if ( d1 != d2 ) return false;

    return true;
}

bool HWVolumeManager::isDevice( const std::string &dev )
{
    NWC::FSEntry fe( dev );

    if ( ! fe.entryExists() ) return false;

    if ( fe.isLink() ) {
        if ( S_ISCHR( fe.stat_dest_mode() ) ||
             S_ISBLK( fe.stat_dest_mode() ) ) return true;
    } else {
        if ( S_ISCHR( fe.stat_mode() ) ||
             S_ISBLK( fe.stat_mode() ) ) return true;
    }

    return false;
}

std::list<HWVolume> HWVolumeManager::getNewVolumes( bool store_new )
{
    std::list<HWVolume> l = getVolumes();
    std::list<HWVolume> newl;
    std::set< std::string > new_known;
    
    for ( std::list<HWVolume>::iterator it1 = l.begin();
          it1 != l.end();
          it1++ ) {
        if ( ! it1->getUDI().empty() ||
             it1->isAvailable() == true ) {
            if ( m_known_devices.count( it1->getDevice() ) == 0 ) {
                newl.push_back( *it1 );
            }
            new_known.insert( it1->getDevice() );
        }
    }
    if ( store_new ) {
        m_known_devices = new_known;
    }
    return newl;
}

std::list<HWVolume> HWVolumeManager::filterUniqueDevices( const std::list<HWVolume> &l ) const
{
    std::set< w_osdevice_t > managed;
    std::list<HWVolume> newl;

    for ( std::list<HWVolume>::const_iterator it1 = l.begin();
          it1 != l.end();
          it1++ ) {
        if ( managed.count( getOSDevice( it1->getDevice() ) ) < 1 ) {
            newl.push_back( *it1 );
            managed.insert( getOSDevice( it1->getDevice() ) );
        }
    }
    return newl;
}

bool HWVolumeManager::getBoolProperty( const std::string &udi,
                                       const std::string &property )
{
#ifdef HAVE_HAL_DBUS
    if ( ! m_hal_ctx ) throw 1;

    if ( ! libhal_device_property_exists( m_hal_ctx, udi.c_str(),
                                          property.c_str(),
                                          NULL ) ) {
        throw 2;
    }

    return libhal_device_get_property_bool( m_hal_ctx, udi.c_str(),
                                            property.c_str(), NULL );
#else
    throw -1;
#endif
}

std::string HWVolumeManager::getStringProperty( const std::string &udi,
                                                const std::string &property )
{
#ifdef HAVE_HAL_DBUS
    if ( ! m_hal_ctx ) throw 1;

    if ( ! libhal_device_property_exists( m_hal_ctx, udi.c_str(),
                                          property.c_str(),
                                          NULL ) ) {
        throw 2;
    }

    char *s1 = libhal_device_get_property_string( m_hal_ctx, udi.c_str(),
                                                  property.c_str(), NULL );
    if ( ! s1 ) throw 3;

    std::string s2( s1 );
    libhal_free_string( s1 );
    return s2;
#else
    throw -1;
#endif
}

std::list< std::string > HWVolumeManager::getStrlistProperty( const std::string &udi,
                                                              const std::string &property )
{
#ifdef HAVE_HAL_DBUS
    if ( ! m_hal_ctx ) throw 1;

    if ( ! libhal_device_property_exists( m_hal_ctx, udi.c_str(),
                                          property.c_str(),
                                          NULL ) ) {
        throw 2;
    }

    char **l1 = libhal_device_get_property_strlist( m_hal_ctx, udi.c_str(),
                                                   property.c_str(), NULL );
    if ( ! l1 ) throw 3;

    std::list< std::string > l2;
    for ( int i = 0; l1[i] != NULL; i++ ) {
        l2.push_back( l1[i] );
    }

    libhal_free_string_array( l1 );

    return l2;
#else
    throw -1;
#endif
}

void HWVolumeManager::setMountCommand( const std::string &cmd )
{
    if ( ! cmd.empty() ) {
        m_mount_command = cmd;
    } else {
        m_mount_command = HW_VM_MOUNT_CMD;
    }
}

void HWVolumeManager::setUnmountCommand( const std::string &cmd )
{
    if ( ! cmd.empty() ) {
        m_unmount_command = cmd;
    } else {
        m_unmount_command = HW_VM_UNMOUNT_CMD;
    }
}

int HWVolumeManager::halMount( HWVolume &vol, std::string &error_return )
{
    if ( vol.isMounted() == true ) {
        error_return = catalog.getLocale( 889 );
        return 1;
    }
    if ( vol.getUDI().empty() || vol.isInFStab() ) {
        error_return = catalog.getLocale( 890 );
        return 1;
    }

#ifdef HAVE_HAL_DBUS
    DBusMessage *message = NULL;
    DBusMessage *reply = NULL;
    const char *mp = "";
    const char *fs = "";
    DBusError error;
#endif
    bool mount_failed = true;

#ifdef HAVE_HAL_DBUS
    if ( HALreconnect() == 0 ) {

        dbus_error_init( &error );

        // check whether there is a mount method
        std::list< std::string > method_list;

        try {
            method_list = getStrlistProperty( vol.getUDI().c_str(), "org.freedesktop.Hal.Device.Volume.method_names" );
        } catch ( int ) {
        }

        if ( std::find( method_list.begin(),
                        method_list.end(),
                        "Mount" ) != method_list.end() ) {

            try {
                message = dbus_message_new_method_call( "org.freedesktop.Hal", 
                                                        vol.getUDI().c_str(),
                                                        "org.freedesktop.Hal.Device.Volume",
                                                        "Mount" );
                if ( ! message ) {
                    //TODO translate
                    error_return = "Mount failed: Unable to create DBus message";
                    throw 1;
                }

                std::list< std::string > valid_options;
                try {
                    valid_options = getStrlistProperty( vol.getUDI().c_str(),
                                                        "volume.mount.valid_options" );
                } catch ( int ) {
                }

                const char *options[] = { NULL };
                int num_options = 0;
                std::string uid_opt;

                if ( std::find( valid_options.begin(),
                                valid_options.end(),
                                "uid=" ) != valid_options.end() ) {
                    uid_opt = AGUIXUtils::formatStringToString( "uid=%d", (int)getuid() );
                    
                    options[0] = uid_opt.c_str();
                    num_options = 1;
                }

                const char **options_p = options;
            
                if ( ! dbus_message_append_args( message, 
                                                 DBUS_TYPE_STRING, &mp,
                                                 DBUS_TYPE_STRING, &fs,
                                                 DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options_p, num_options,
                                                 DBUS_TYPE_INVALID ) ) {
                    //TODO translate?
                    error_return = "Mount failed: Unable to add arguments to DBus message";
                    throw 2;
                }
            
                reply = dbus_connection_send_with_reply_and_block( m_dbus_conn,
                                                                   message,
                                                                   -1,
                                                                   &error );
                if ( ! reply ) {
                    throw 3;
                }

                if ( dbus_error_is_set( &error ) ) {
                    throw 4;
                }
            
#if 0
                int hal_retcode;

                // I do not do this test anymore, on some systems its INT32
                // on other UINT32
                // beside, other tools (gnome-mount) do not check it either
                if ( ! dbus_message_get_args( reply, 
                                              &error,
                                              DBUS_TYPE_UINT32, &hal_retcode,
                                              DBUS_TYPE_INVALID ) ) {
                    throw 4;
                }
            
                if ( hal_retcode != 0 ) {
                    throw 5;
                }
#endif
                mount_failed = false;
            } catch ( int e ) {
                if ( dbus_error_is_set( &error ) ) {
                    error_return = AGUIXUtils::formatStringToString( catalog.getLocale( 891 ),
                                                                     error.name,
                                                                     error.message );
                    dbus_error_free( &error );
                }
            }

            if ( message ) dbus_message_unref( message );
            if ( reply ) dbus_message_unref( reply );
        } else {
            // no mount method
            //TODO translate?
            error_return = "Mount failed: No Mount method";
        }
    }
#endif
    if ( mount_failed ) return 1;

    update( vol );

    return 0;
}

int HWVolumeManager::cmdMount( HWVolume &vol, std::string &error_return )
{
    if ( vol.isMounted() == true ) {
        error_return = catalog.getLocale( 889 );
        return 1;
    }
    if ( ! vol.getUDI().empty() && vol.isInFStab() == false ) {
        error_return = catalog.getLocale( 892 );
        return 1;
    }

    FlagReplacer::FlagProducerMap fl;

    if ( ! vol.getSupposedMountPoint().empty() ) {
        fl.registerFlag( "f", vol.getSupposedMountPoint() );
    } else {
        fl.registerFlag( "f", vol.getDevice() );
    }
    fl.registerFlag( "d", vol.getDevice() );

    FlagReplacer rpl( fl );
  
    std::string exestr = rpl.replaceFlags( m_mount_command );

    ExeClass ec( ExeClass::EXECUTE_NORMAL );

    ec.addCommand( "%s", exestr.c_str() );
    int ret = ec.getReturnCode();
    if ( ret != 0 ) {
        std::string error;
        //TODO do not use a hard-coded limit
        ec.readErrorOutput( error, 1024 );

        error_return = error.substr( 0, error.find( '\n' ) );
        return 1;
    }

    update( vol );

    return 0;
}

int HWVolumeManager::halUnmount( HWVolume &vol, std::string &error_return )
{
    if ( vol.isMounted() == false ) {
        error_return = catalog.getLocale( 893 );
        return 1;
    }
    if ( vol.getUDI().empty() || vol.isInFStab() ) {
        error_return = catalog.getLocale( 894 );
        return 1;
    }

#ifdef HAVE_HAL_DBUS
    DBusMessage *message = NULL;
    DBusMessage *reply = NULL;
    DBusError error;
#endif
    bool unmount_failed = true;
    
#ifdef HAVE_HAL_DBUS
    if ( HALreconnect() == 0 ) {

        dbus_error_init( &error );

        // check whether there is a mount method
        std::list< std::string > method_list;

        try {
            method_list = getStrlistProperty( vol.getUDI().c_str(), "org.freedesktop.Hal.Device.Volume.method_names" );
        } catch ( int ) {
        }

        if ( std::find( method_list.begin(),
                        method_list.end(),
                        "Unmount" ) != method_list.end() ) {

            try {
                message = dbus_message_new_method_call( "org.freedesktop.Hal", 
                                                        vol.getUDI().c_str(),
                                                        "org.freedesktop.Hal.Device.Volume",
                                                        "Unmount" );
                if ( ! message ) {
                    //TODO translate?
                    error_return = "Unmount failed: Unable to create DBus message";
                    throw 1;
                }

                const char *options[] = { NULL };
                int num_options = 0;
                const char **options_p = options;
            
                if ( ! dbus_message_append_args( message, 
                                                 DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options_p, num_options,
                                                 DBUS_TYPE_INVALID ) ) {
                    //TODO translate?
                    error_return = "Unmount failed: Unable to add arguments to DBus message";
                    throw 2;
                }
            
                reply = dbus_connection_send_with_reply_and_block( m_dbus_conn,
                                                                   message,
                                                                   -1,
                                                                   &error );
                if ( ! reply ) {
                    throw 3;
                }
            
                if ( dbus_error_is_set( &error ) ) {
                    throw 4;
                }

#if 0
                int hal_retcode;

                // see comment in halMount
                if ( ! dbus_message_get_args( reply, 
                                              &error,
                                              DBUS_TYPE_UINT32, &hal_retcode,
                                              DBUS_TYPE_INVALID ) ) {
                    throw 4;
                }
            
                if ( hal_retcode != 0 ) {
                    throw 5;
                }
#endif
                unmount_failed = false;
            } catch ( int e ) {
                if ( dbus_error_is_set( &error ) ) {
                    error_return = AGUIXUtils::formatStringToString( catalog.getLocale( 895 ),
                                                                     error.name,
                                                                     error.message );
                    dbus_error_free( &error );
                }
            }

            if ( message ) dbus_message_unref( message );
            if ( reply ) dbus_message_unref( reply );
        } else {
            // no unmount method
            //TODO translate?
            error_return = "Unmount failed: No Unmount method";
        }
    }
#endif
    if ( unmount_failed ) return 1;

    update( vol );

    return 0;
}

int HWVolumeManager::cmdUnmount( HWVolume &vol, std::string &error_return )
{
    if ( vol.isMounted() == false ) {
        error_return = catalog.getLocale( 893 );
        return 1;
    }
    if ( ! vol.getUDI().empty() && vol.isInFStab() == false ) {
        error_return = catalog.getLocale( 896 );
        return 1;
    }

    FlagReplacer::FlagProducerMap fl;

    if ( ! vol.getSupposedMountPoint().empty() ) {
        fl.registerFlag( "f", vol.getSupposedMountPoint() );
    } else {
        fl.registerFlag( "f", vol.getDevice() );
    }
    fl.registerFlag( "d", vol.getDevice() );

    FlagReplacer rpl( fl );
  
    std::string exestr = rpl.replaceFlags( m_unmount_command );

    ExeClass ec( ExeClass::EXECUTE_NORMAL );

    ec.addCommand( "%s", exestr.c_str() );
    int ret = ec.getReturnCode();
    if ( ret != 0 ) {
        std::string error;
        //TODO do not use a hard-coded limit
        ec.readErrorOutput( error, 1024 );

        error_return = error.substr( 0, error.find( '\n' ) );
        return 1;
    }

    update( vol );

    return 0;
}

void HWVolumeManager::setPartitionFile( const std::string &file )
{
    if ( ! file.empty() ) {
        m_partition_file = file;
    } else {
        m_partition_file = HW_VM_PARTS;
    }
}

void HWVolumeManager::setFStabFile( const std::string &file )
{
    if ( ! file.empty() ) {
        m_fstab_table = FStabMountTable( file );
    } else {
        m_fstab_table = FStabMountTable( HW_VM_FSTAB );
    }
}

void HWVolumeManager::setMtabFile( const std::string &file )
{
    if ( ! file.empty() ) {
        m_mtab_table = MtabMountTable( file );
    } else {
        m_mtab_table = MtabMountTable( HW_VM_MTAB );
    }
}

int HWVolumeManager::HALconnectionInit()
{
#ifdef HAVE_HAL_DBUS
    bool init_failed = false;
    DBusError error;

    if ( HALisConnected() ) return 0;

    dbus_error_init( &error );

    m_hal_ctx = libhal_ctx_new();
    if ( m_hal_ctx != NULL ) {
        
        m_dbus_conn = dbus_bus_get( DBUS_BUS_SYSTEM, &error );
        
        if ( dbus_error_is_set( &error ) ) {
            fprintf( stderr, "DBUS: error %s\n", error.message );
            dbus_error_free( &error );
            init_failed = true;
        } else {
            libhal_ctx_set_dbus_connection( m_hal_ctx,
                                            m_dbus_conn );

            if ( ! libhal_ctx_init( m_hal_ctx, &error ) ) {
		if ( dbus_error_is_set( &error ) ) {
                    fprintf( stderr, "libhal_ctx_init: %s: %s\n", error.name, error.message );
                    dbus_error_free( &error );
		} else {
                    fprintf( stderr, "libhal_ctx_init failed\n" );
		}
                init_failed = true;
            }
        }
    } else {
        init_failed = true;
    }

    if ( init_failed == true ) {
        if ( m_hal_ctx != NULL ) {
            libhal_ctx_free( m_hal_ctx );
            m_hal_ctx = NULL;
        }
        if ( m_dbus_conn != NULL ) {
            dbus_connection_unref( m_dbus_conn );
            m_dbus_conn = NULL;
        }
    }
    return ( init_failed == true ) ? 1 : 0;
#else
    return 0;
#endif
}

int HWVolumeManager::HALconnectionFini()
{
#ifdef HAVE_HAL_DBUS
    DBusError error;

    dbus_error_init( &error );

    if ( m_hal_ctx != NULL ) {
        if ( ! libhal_ctx_shutdown( m_hal_ctx, &error ) ) {
            if ( dbus_error_is_set( &error ) ) {
                fprintf( stderr, "libhal_ctx_shutdown: %s: %s\n", error.name, error.message );
                dbus_error_free( &error );
            } else {
                fprintf( stderr, "libhal_ctx_shutdown failed.\n" );
            }
        }
        
        libhal_ctx_free( m_hal_ctx );
    }
    if ( m_dbus_conn != NULL ) {
        dbus_connection_unref( m_dbus_conn );
    }
    return 0;
#else
    return 0;
#endif
}

bool HWVolumeManager::HALisConnected()
{
#ifdef HAVE_HAL_DBUS
    if ( m_dbus_conn != NULL ) {
        if ( dbus_connection_get_is_connected( m_dbus_conn ) ) {
            return true;
        }
    }
    return false;
#else
    return false;
#endif
}

int HWVolumeManager::HALreconnect()
{
#ifdef HAVE_HAL_DBUS
    if ( HALisConnected() ) return 0;

    if ( HALconnectionFini() != 0 ) return 1;

    if ( HALconnectionInit() != 0 ) return 1;

    return 0;
#else
    return 0;
#endif
}
